/*
 *  $Id: bgp_sm.c,v 1.5 1998/06/03 21:38:10 masaki Exp $
 */

#include <version.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>

#include <mrt.h>
#include <timer.h>
#include <trace.h>
#include <select.h>
#include <bgp.h>


/* bgp_change_state
 * A convenience routine that does tracing on state
 * transition.
 */
void 
bgp_change_state (bgp_peer_t * peer, int state, int event)
{
    /* should not happen ? */
    if (peer->state == state)
	return;

    trace (TR_STATE, peer->trace, "%s -> %s (event %s)\n",
	   sbgp_states[peer->state], sbgp_states[state], sbgp_events[event]);

    if (BGP->state_change_fn)
        BGP->state_change_fn (peer, state);

    peer->state = state;
    time (&peer->time);

    if (state == BGPSTATE_ESTABLISHED) {
	if (peer->sockfd_in >= 0) {
            trace (TR_STATE, peer->trace, "closing the incoming connection\n");
	    select_delete_fd (peer->sockfd_in);
	    peer->sockfd_in = -1;
	}
    }

    if (state == BGPSTATE_ESTABLISHED && BGP->peer_established_call_fn)
	BGP->peer_established_call_fn (peer);

    if (state == BGPSTATE_IDLE && BIT_TEST (BGP_CONNECT_PASSIVE, peer->options))
	bgp_change_state (peer, BGPSTATE_ACTIVE, BGPEVENT_START);
}


void 
bgp_sm_state_idle (bgp_peer_t * peer, int event)
{
    /* ignore all events but START event */
    /* connection retry timeout */
    if (event != BGPEVENT_START && event != BGPEVENT_CONNRETRY) {
	trace (TR_STATE, peer->trace, "Idle -> Idle (event %s IGNORED)\n",
	       sbgp_events[event]);
	return;
    }

    Timer_Turn_ON (peer->timer_ConnectRetry);

    trace (TR_STATE, peer->trace, "Attempting connection\n");

    bgp_change_state (peer, BGPSTATE_CONNECT, BGPEVENT_START);
    if (bgp_start_transport_connection (peer) < 0) {
	trace (TR_STATE, peer->trace, "connection failed (%s)\n",
	       strerror (errno));
	bgp_change_state (peer, BGPSTATE_ACTIVE, BGPEVENT_OPENFAIL);
    }
}



void 
bgp_sm_state_active (bgp_peer_t * peer, int event)
{
    switch (event) {
    case BGPEVENT_START:
	return;			/* do nothing */
	break;
    case BGPEVENT_OPEN:
	if (!BIT_TEST (BGP_CONNECT_PASSIVE, peer->options))
	    Timer_Turn_OFF (peer->timer_ConnectRetry);
	select_add_fd (peer->sockfd, 1, bgp_schedule_socket, peer);
	bgp_change_state (peer, BGPSTATE_OPENSENT, event);
	bgp_send_open (peer);
	break;
    case BGPEVENT_OPENFAIL:
	if (!BIT_TEST (BGP_CONNECT_PASSIVE, peer->options))
	    Timer_Reset_Time (peer->timer_ConnectRetry);
	break;
    case BGPEVENT_CONNRETRY:
	Timer_Reset_Time (peer->timer_ConnectRetry);
	bgp_change_state (peer, BGPSTATE_CONNECT, event);
	if (bgp_start_transport_connection (peer) < 0) {
	    trace (TR_INFO, peer->trace, "Transport connection failed (%s)\n",
		   strerror (errno));
	    bgp_sm_process_event (peer, BGPEVENT_OPENFAIL);
	}
	break;
    default:
	bgp_change_state (peer, BGPSTATE_IDLE, event);
	bgp_peer_dead (peer);
	break;
    }
}

void 
bgp_sm_state_connect (bgp_peer_t * peer, int event)
{
    if (peer->sockfd >= 0)
	select_delete_fd2 (peer->sockfd);

    switch (event) {
    case BGPEVENT_START:
	return;			/* do nothing */
	break;
    case BGPEVENT_OPEN:
	if (!BIT_TEST (BGP_CONNECT_PASSIVE, peer->options))
	    Timer_Turn_OFF (peer->timer_ConnectRetry);
	select_add_fd (peer->sockfd, 1, bgp_schedule_socket, peer);
	bgp_change_state (peer, BGPSTATE_OPENSENT, event);
	bgp_send_open (peer);
	break;
    case BGPEVENT_OPENFAIL:
	if (!BIT_TEST (BGP_CONNECT_PASSIVE, peer->options))
	    Timer_Reset_Time (peer->timer_ConnectRetry);
	bgp_change_state (peer, BGPSTATE_ACTIVE, event);
	break;
    case BGPEVENT_CONNRETRY:
	Timer_Reset_Time (peer->timer_ConnectRetry);
	bgp_change_state (peer, BGPSTATE_CONNECT, event);
	if (bgp_start_transport_connection (peer) < 0) {
	    trace (TR_INFO, peer->trace, "Transport connection failed (%s)\n",
		   strerror (errno));
	    bgp_sm_process_event (peer, BGPEVENT_OPENFAIL);
	}
	break;
    default:
	bgp_change_state (peer, BGPSTATE_IDLE, event);
	bgp_peer_dead (peer);
	break;
    }
}


void 
bgp_sm_state_opensent (bgp_peer_t * peer, int event)
{

    switch (event) {
    case BGPEVENT_START:
	return;			/* do nothing */
	break;
    case BGPEVENT_RECVNOTIFY:
    case BGPEVENT_CLOSED:
	/* close transport connection */
	if (!BIT_TEST (BGP_CONNECT_PASSIVE, peer->options)) {
	    if (peer->sockfd >= 0) {
		select_delete_fd (peer->sockfd);
		peer->sockfd = -1;
	    }
	    Timer_Reset_Time (peer->timer_ConnectRetry);
	}
	bgp_change_state (peer, BGPSTATE_ACTIVE, event);
	break;
    case BGPEVENT_ERROR:
	bgp_change_state (peer, BGPSTATE_IDLE, event);
	bgp_peer_dead (peer);
	break;
    case BGPEVENT_RECVOPEN:
	/* if OPEN is good */
	if (bgp_process_open (peer) > 0) {
	    bgp_change_state (peer, BGPSTATE_OPENCONFIRM, event);
	    bgp_send_keepalive (peer);
	}
	/* OPEN was bad */
	else {
	    /* notification already sent */
	    bgp_change_state (peer, BGPSTATE_IDLE, event);
	    bgp_peer_dead (peer);
	}
	break;
    default:
	/* send notification */
	/* close transport */
	bgp_change_state (peer, BGPSTATE_IDLE, event);
	bgp_peer_dead (peer);
	break;
    }
}


void 
bgp_sm_state_openconfirm (bgp_peer_t * peer, int event)
{

    switch (event) {
    case BGPEVENT_START:
	return;			/* do nothing */
	break;
    case BGPEVENT_CLOSED:
    case BGPEVENT_OPENFAIL:
	bgp_change_state (peer, BGPSTATE_IDLE, event);
	bgp_release_resources (peer);
	break;
    case BGPEVENT_KEEPALIVE:
	Timer_Reset_Time (peer->timer_KeepAlive);
	bgp_send_keepalive (peer);
	break;
    case BGPEVENT_RECVKEEPALIVE:
	Timer_Reset_Time (peer->timer_HoldTime);
	bgp_change_state (peer, BGPSTATE_ESTABLISHED, event);
	bgp_establish_peer (peer);
	break;
    case BGPEVENT_RECVNOTIFY:
	/* close transport connection */
	bgp_change_state (peer, BGPSTATE_IDLE, event);
	bgp_peer_dead (peer);
	break;
    default:
	/* send notification */
	bgp_change_state (peer, BGPSTATE_IDLE, event);
	bgp_peer_dead (peer);
	break;
    }
}



void 
bgp_sm_state_established (bgp_peer_t * peer, int event)
{

    switch (event) {
    case BGPEVENT_START:
	return;			/* do nothing */
	break;
    case BGPEVENT_CLOSED:
    case BGPEVENT_ERROR:
	bgp_change_state (peer, BGPSTATE_IDLE, event);
	/* view_delete_peer (BGP->view, peer->gateway); */
	bgp_peer_dead (peer);
	break;
    case BGPEVENT_KEEPALIVE:
	Timer_Reset_Time (peer->timer_KeepAlive);
	bgp_send_keepalive (peer);
	break;
    case BGPEVENT_RECVKEEPALIVE:
	Timer_Reset_Time (peer->timer_HoldTime);
	break;
    case BGPEVENT_RECVUPDATE:
	Timer_Reset_Time (peer->timer_HoldTime);
	bgp_process_update (peer);
	/* check update PDU */
	break;
    case BGPEVENT_RECVNOTIFY:
	bgp_change_state (peer, BGPSTATE_IDLE, event);
	bgp_peer_dead (peer);
	break;
    default:
	/* send notification */
	/* we may need to check BGP_ERR_FSM */
	bgp_change_state (peer, BGPSTATE_IDLE, event);
	bgp_peer_dead (peer);
	break;
    }
}


void 
bgp_sm_process_event (bgp_peer_t * peer, int event)
{
    assert (peer);

    switch (peer->state) {
    case BGPSTATE_IDLE:
	bgp_sm_state_idle (peer, event);
	break;
    case BGPSTATE_ACTIVE:
	bgp_sm_state_active (peer, event);
	break;
    case BGPSTATE_CONNECT:
	bgp_sm_state_connect (peer, event);
	break;
    case BGPSTATE_OPENSENT:
	bgp_sm_state_opensent (peer, event);
	break;
    case BGPSTATE_OPENCONFIRM:
	bgp_sm_state_openconfirm (peer, event);
	break;
    case BGPSTATE_ESTABLISHED:
	bgp_sm_state_established (peer, event);
	break;
    default:
	assert (0);
	break;
    }
}
